@@ -13,7 +13,7 @@ use Math::BigInt try => 'GMP,Pari';
 use Exporter 'import';
 our @EXPORT_OK = qw(b64url_encode b64url_decode);
 
-our $VERSION = '0.01';
+our $VERSION = '0.02';
 our $DEBUG = 0;
 our $GENERATOR;
 
@@ -24,6 +24,7 @@ our $MAX_ROUNDS = 100;
 sub random_nbit_prime;
 sub prime_set_config;
 
+
 # Load Math::Prime::Util and Math::Random::Secure
 BEGIN {
   if (eval
@@ -47,7 +48,8 @@ sub new {
   return $_[0] if ref $_[0] && ref $_[0] eq __PACKAGE__;
 
   # MagicKey in string notation
-  if (@_ == 1 && $_[0] =~ /rsa/i ) {
+  if (@_ == 1 && index($_[0], 'RSA') >= 0) {
+
     my $string = shift;
     return unless $string;
 
@@ -64,7 +66,7 @@ sub new {
     my ($type, $mod, $exp, $private_exp) = split(/\./, $string);
 
     # The key is incorrect
-    if (uc($type) ne 'RSA') {
+    if ($type ne 'RSA') {
       carp "MagicKey type '$type' is not supported";
       return;
     };
@@ -92,12 +94,12 @@ sub new {
     # RSA complete description
     if (defined $param{n}) {
 
-      my %self;
-      foreach (qw/n e d/) {
-	$self{$_} = Math::BigInt->new($param{$_}) if exists $param{$_};
-      };
+      $self = bless {}, $class;
 
-      $self = bless \%self, $class;
+      # Set attributes
+      foreach (qw/n e d/) {
+	$self->$_($param{$_}) if exists $param{$_};
+      };
 
       unless ($self->n) {
 	carp 'Key is not well defined';
@@ -118,7 +120,7 @@ sub new {
       my $size = $param{size} || 512;
 
       # Key size is too short or impractical
-      return undef if $size < 512 || $size % 2;
+      return undef if $size < 512 || $size > 1024 || $size % 2;
 
       # Public exponent
       my $e = $param{e};
@@ -167,13 +169,14 @@ sub new {
       # Bless object
       $self = bless {}, $class;
 
+      # Set e
       $self->e($e) if $e;
 
       # Calculate phi
       my $phi = ($p - 1) * ($q - 1);
 
       # Calculate multiplicative inverse of e modulo phi
-      my $d = Math::BigInt->new($self->e)->bmodinv($phi);
+      my $d = $self->e->copy->bmodinv($phi);
 
       # $d is too short
       goto CALC_KEY if _bitsize($d) < $size / 4;
@@ -186,12 +189,14 @@ sub new {
     };
   };
 
-  # Set emLen (octet length of modulus)
-  $self->{emLen} = _octet_len( $self->n );
+  # Todo: Check that n is not tooo large!
 
   # Set size (bitsize length of modulus)
   $self->{size} = _bitsize( $self->n );
 
+  # Set emLen (octet length of modulus)
+  $self->{emLen} = _octet_len( $self->n );
+
   return $self;
 };
 
@@ -202,17 +207,21 @@ sub n {
 
   # Get value
   unless ($_[0]) {
-    return $self->{n} // 0;
+    return ($self->{n} //= Math::BigInt->bzero);
   }
 
   # Set value
-  elsif ($_[0] =~ /^\d+$/) {
+  else {
+    my $n = Math::BigInt->new( shift );
+
+    # n is not a number
+    carp 'n is not a number' and return undef if $n->is_nan;
 
     # Delete precalculated emLen and size
     delete $self->{emLen};
     delete $self->{size};
 
-    return $self->{n} = Math::BigInt->new( shift );
+    return $self->{n} = $n;
   };
 
   return undef;
@@ -225,12 +234,17 @@ sub e {
 
   # Get value
   unless ($_[0]) {
-    return $self->{e} // 65537;
+    return ($self->{e} //= Math::BigInt->new('65537'));
   }
 
   # Set value
-  elsif ($_[0] =~ /^\d+$/) {
-    return $self->{e} = Math::BigInt->new( shift );
+  else {
+    my $e = Math::BigInt->new( shift );
+
+    # e is not a number
+    carp 'e is not a number' and return undef if $e->is_nan;
+
+    return $self->{e} = $e;
   };
 
   return undef;
@@ -243,12 +257,17 @@ sub d {
 
   # Get value
   unless ($_[0]) {
-    return $self->{d} // 0;
+    return $self->{d} // undef;
   }
 
   # Set value
-  elsif ($_[0] =~ /^\d+$/) {
-    return $self->{d} = Math::BigInt->new( shift );
+  else {
+    my $d = Math::BigInt->new( shift );
+
+    # d is not a number
+    carp 'd is not a number' and return undef if $d->is_nan;
+
+    return $self->{d} = $d;
   };
 
   return undef;
@@ -490,13 +509,15 @@ sub _emsa_encode {
 
 
 # Convert from octet string to bigint
-sub _os2ip ($) {
+sub _os2ip {
   # Based on Crypt::RSA::DataFormat
 
   my $os = shift;
+  my $l = length($os);
+  return undef if $l > 30_000;
+
   my $base = Math::BigInt->new(256);
   my $result = 0;
-  my $l = length($os);
   for (0 .. $l-1) {
     my ($c) = unpack "x$_ a", $os;
     my $a = int(ord($c));
@@ -505,7 +526,7 @@ sub _os2ip ($) {
     # Maybe optimizable
     $result += $a * ($base**$val);
   };
-  return $result;
+  $result;
 };
 
 
@@ -513,7 +534,11 @@ sub _os2ip ($) {
 sub _i2osp {
   # Based on Crypt::RSA::DataFormat
 
-  my $num = Math::BigInt->new( shift );
+  my $num = Math::BigInt->new(shift);
+
+  return if $num->is_nan;
+  return if $num->length > 30_000;
+
   my $l = shift || 0;
 
   my $result = '';
@@ -535,7 +560,7 @@ sub _i2osp {
     $result = chr(0) x ($l - length($result)) . $result;
   };
 
-  return $result;
+  $result;
 };
 
 
@@ -552,7 +577,7 @@ sub _octet_len {
 sub _bitsize {
   my $int = Math::BigInt->new( shift );
   return 0 unless $int;
-  return ( length( $int->as_bin ) - 2 );
+  length( $int->as_bin ) - 2;
 };
 
 
@@ -580,18 +605,16 @@ sub _hex_to_b64url {
 
 # Returns the b64 urlsafe encoding of a string
 sub b64url_encode ($;$) {
-  my $v = shift;
-  my $p = defined $_[0] ? shift : 1;
+  return '' unless $_[0];
 
-  return '' unless $v;
+  my $v = $_[0];
 
   utf8::encode $v if utf8::is_utf8 $v;
   $v = encode_base64($v, '');
-  $v =~ tr{+/}{-_};
-  $v =~ tr{\t-\x0d }{}d;
+  $v =~ tr{+/\t-\x0d }{-_}d;
 
   # Trim padding or not
-  $v =~ s/\=+$// unless $p;
+  $v =~ s/\=+$// unless (defined $_[1] ? $_[1] : 1);
 
   return $v;
 };
@@ -604,8 +627,10 @@ sub b64url_decode ($) {
 
   $v =~ tr{-_}{+/};
 
+  my $padding;
+
   # Add padding
-  if (my $padding = (length($v) % 4)) {
+  if ($padding = (length($v) % 4)) {
     $v .= chr(61) x (4 - $padding);
   };
 
@@ -650,7 +675,7 @@ B<This module is an early release! There may be significant changes in the futur
 =head1 ATTRIBUTES
 
 
-=head2 C<n>
+=head2 n
 
   print $mkey->n;
   $mkey->n('456789...');
@@ -658,7 +683,7 @@ B<This module is an early release! There may be significant changes in the futur
 The MagicKey modulus.
 
 
-=head2 C<e>
+=head2 e
 
   print $mkey->e;
   $mkey->e(3);
@@ -667,7 +692,7 @@ The MagicKey public exponent.
 Defaults to C<65537>.
 
 
-=head2 C<d>
+=head2 d
 
   print $mkey->d;
   $mkey->d('234567...');
@@ -675,7 +700,7 @@ Defaults to C<65537>.
 The MagicKey private exponent.
 
 
-=head2 C<size>
+=head2 size
 
   print $mkey->size;
 
@@ -684,7 +709,7 @@ The MagicKey keysize in bits.
 
 =head1 METHODS
 
-=head2 C<new>
+=head2 new
 
   my $mkey = Crypt::MagicSignatures::Key->new(<<'MKEY');
     RSA.
@@ -717,7 +742,7 @@ In case no C<size> attribute is given, the default key size
 for generation is 512 bits.
 
 
-=head2 C<sign>
+=head2 sign
 
   my $sig = $mkey->sign('This is a message');
 
@@ -727,7 +752,7 @@ The signature algorithm is based on
 L<RFC3447|http://www.ietf.org/rfc/rfc3447.txt>.
 
 
-=head2 C<verify>
+=head2 verify
 
   my $sig = $priv_key->sign('This is a message');
 
@@ -746,7 +771,7 @@ component of the key.
 Returns a C<true> value on success and C<false> otherwise.
 
 
-=head2 C<to_string>
+=head2 to_string
 
   my $pub_key = $mkey->to_string;
   my $priv_key = $mkey->to_string(1);
@@ -762,7 +787,7 @@ is returned.
 
 =head1 FUNCTIONS
 
-=head2 C<b64url_encode>
+=head2 b64url_encode
 
   use Crypt::MagicSignatures::Key qw/b64url_encode/;
 
@@ -777,7 +802,7 @@ L<encode_base64 in MIME::Base64|MIME::Base64/"encode_base64">.
 The function can be exported.
 
 
-=head2 C<b64url_decode>
+=head2 b64url_decode
 
   use Crypt::MagicSignatures::Key qw/b64url_decode/;
 
@@ -799,8 +824,9 @@ L<MIME::Base64>.
 L<Math::Prime::Util> and
 L<Math::Random::Secure> are necessary for key generation only.
 
-Either L<Math::BigInt::GMP> (preferred) or L<Math::BigInt::Pari> are recommended
-for speed, as well as L<Math::Random::ISAAC::XS>.
+Either L<Math::BigInt::GMP> (preferred) or L<Math::BigInt::Pari>
+are strongly recommended for speed,
+as well as L<Math::Random::ISAAC::XS>.
 
 
 =head1 KNOWN BUGS AND LIMITATIONS
